//
// Module: RegionFunctions
// Author: Richard Saxton
//
// Modifications: See ChangeLog
//
// Purpose: Contains methods which manipulate the DSS via regions
//
// General classes
#include "Tracer.h"

/* SAS-specific */
#include "DataFilter.h"
#include "Region.h"
#include "DataSubspace.h"
#include "ASCregion.h"
#include "Dal.h"
#include "common.h"
#include "OalServer.h"


/* Local classes */
#include "SourcePosition.h"
#include "RegionFunctions.h"
#include "CoreData.h"
#include "PositionUtils.h"

/* Constructor */
RegionFunctions::RegionFunctions(CoreData * coreData) :
_coredata(coreData)
{
  PROC;

  /* Set the position utilities object */
    _putils = _coredata->varpositionUtils();

  ENDPROC;
}

/* Destructor */
RegionFunctions::~RegionFunctions(){};

/* Method region(DataSubspace dss) - convert a dss filter into a r
egion */
GeoRegion RegionFunctions::region(DataSubspace * dss) 
{
   PROC;

   RETURN region(dss, false);
}

/* Method region(DataSubspace * dss, bool rawonly) - convert a dss filter into a region */
GeoRegion RegionFunctions::region(DataSubspace * dss, bool rawonly) 
{
   PROC;

/*  If rawonly=true create the region for RAWX/Y filters only. 
    Otherwise ignore RAWX/Y filters */

  DSS::DSaxes axes;
  axes.push_back("RAW");
  string projregionexpr = dss->selectExpression(axes);
  errstr << "Expression 2 "<<projregionexpr
         << Sas::message(Msg::App, Msg::Verbose);


    GeoRegion region;

   /* Set a boolean to decide whether the region contains X/Y filters */
    _regPosCoords = false;

   /* Get the filter corresponding to the region block within the dss */
    for (unsigned i  = 0; i < dss->components(); ++i) {

      DScomp * dscomp = dss->component(i);

     /* Loop over each filter in this component */
      for (unsigned f = 0; f < dscomp->filters(); ++f) {

          DSfilter * dsfilt = dscomp->filter(f);

          DSS::DSfilterComp caxes = dsfilt->comps();

        /* If the filter is a region - get a region object
           Make sure that global filters are only included once.
           Exclude RAWX/Y regions from RegionT definitions because
           they are usually included needlessly */

           bool globalFilter = dss->isGlobal((*dsfilt));
           if ( (! globalFilter || i==0) &&
                (dsfilt->type() == DSS::RegionBlockT
                               || dsfilt->type() == DSS::RegionT) &&
                ( (rawonly && (caxes[0] == "RAWX" || caxes[1] == "RAWX")) ||
                ( !rawonly && caxes[0] != "RAWX" && caxes[1] != "RAWX") )
              )
           {
              GeoRegion tempregion;

              if (dsfilt->type() == DSS::RegionBlockT) {

               /* Write region spec into a separate region object */
                InASCregionTable ascTab(dsfilt->blockSpec());
                ascTab >> tempregion;

              } else if (dsfilt->type() == DSS::RegionT) {
                tempregion = dsfilt->region();
              }
              for (unsigned comp=0; comp<tempregion.size(); comp++) {
                  region.push_back(tempregion[comp]);
              }

            /* Check whether the filter is in POS coords */
             if ( caxes[0] == "X" || caxes[1] == "X") {
                _regPosCoords = true;
             }

             errstr<<"region block added: " << dsfilt->selectExpression()
               <<" comp="<<i
               << Sas::message(Msg::App, Msg::Verbose);
           }
      }
    }

    errstr<<"regPos= "<<_regPosCoords
          << Sas::message(Msg::App, Msg::Verbose);

    RETURN region;
}


/* Method convRawToDet(DataSubspace dss) - converts any RAW filters to DET */
void RegionFunctions::convRawToDet(DataSubspace * dss)
{
  /* Produce a region containing all the RAW filters from the
     DSS recoded as DET filters */
  
    GeoRegion rawDetReg = rawBlockToDet(region(dss, true));

  /* Add these new filters to the datasubspace */
    _dssextra = new DataSubspace();
    DScomp * dssextracomp = _dssextra->addComponent();
    dssextracomp->addRegionFilter("DET",rawDetReg);

    for (unsigned icomp  = 0; icomp < rawDetReg.size(); ++icomp) {
          dss->addComponent()->addRegionFilter("DET",rawDetReg);

          errstr<<"Added the DET bit"
               << Sas::message(Msg::App, Msg::Verbose);
    }

  /* Now remove the original RAW filters */
    removeFilters(dss, "RAWX");

   ENDPROC;
}
void RegionFunctions::removeFilters(DataSubspace * dss, const string & axis)
{
    for (unsigned i  = 0; i < dss->components(); ++i) {

      DScomp * dscomp = dss->component(i);

     /* Loop over each filter in this component */
      for (unsigned f = 0; f < dscomp->filters(); ++f) {

          DSfilter * dsfilt = dscomp->filter(f);

          DSS::DSfilterComp caxes = dsfilt->comps();

          errstr<<"My region block del " << dsfilt->selectExpression()
               << Sas::message(Msg::App, Msg::Verbose);

          if (caxes[0] == axis || caxes[1] == axis) {
             dscomp->deleteFilter(f);
          }
      }
    }

    ENDPROC;
}


/* Method: rawBlockToDet - converts RAW spatial selection to det coords */
GeoRegion RegionFunctions::rawBlockToDet(GeoRegion inRegion) const
{
  PROC;

   GeoRegion outRegion;

  /* Loop over each component in the input region */
   for (unsigned comp=0; comp<inRegion.size(); comp++) {

      Component c;
      FShape * s;

     /* Loop over each shape in this component */
      for (unsigned element=0; element<inRegion[comp].size(); element++) {

       /* Get RAW coord values for the shape */
        std::vector<float> X = inRegion[comp][element]->pointsX();
        std::vector<float> Y = inRegion[comp][element]->pointsY();
        std::vector<float> R = inRegion[comp][element]->radii();
        float rotangle = inRegion[comp][element]->rotAngles()[0];
        bool included = inRegion[comp][element]->included();

       /* Convert to DET coords */
        std::vector<float> xCoord, yCoord, dradii;

        for (unsigned i  = 0; i < X.size(); ++i) {
            CoordPair dpair = _putils->RawTodetxy005(X[i],Y[i]);
            xCoord.push_back(dpair.x());
            yCoord.push_back(dpair.y());
        } 

       /* Set refpoint */
        Point<float>    refPoint(xCoord[0], yCoord[0]);

       /* Convert radii */
        for (unsigned i  = 0; i < R.size(); ++i) {
           dradii.push_back(R[i] * _putils->detxyWidth()[0]);
          errstr<<"Set radius 1 raw: "<<R[i]<<" det: "<<dradii[i]
                << Sas::message(Msg::App, Msg::Verbose);
        }

      /* Create an output shape the same as the input shape */
       switch(inRegion[comp][element]->type())
        {
         case ShapeBasics::RECTANGLE :
         {
          s = new FRectangle(refPoint, FPoint(xCoord[1], yCoord[1]), included);
          break;
         }
         case ShapeBasics::ROTRECTANGLE :
         {
          s = new FRotRectangle(refPoint, FPoint(xCoord[1], yCoord[1]),
                                   rotangle, included);
          break;
         }
         case ShapeBasics::BOX :
         {
          s = new FBox(refPoint, dradii[0], dradii[1], included);
          break;
         }
         case ShapeBasics::ROTBOX :
         {
          s = new FRotBox(refPoint, dradii[0], dradii[1], 
                                    rotangle, included);
          break;
         }
         case ShapeBasics::RHOMBUS :
         {
          s = new FRhombus(refPoint, dradii[0], dradii[1], included);
          break;
         }
         case ShapeBasics::ROTRHOMBUS :
         {
          s = new FRotRhombus(refPoint, dradii[0], dradii[1], 
                                    rotangle, included);
          break;
         }
         case ShapeBasics::POLYGON :
         {
          s = new FPolygon(xCoord, yCoord, xCoord.size(), included);
          break;
         }
         case ShapeBasics::CIRCLE: 
         {
          s = new FCircle(refPoint, dradii[0], included);
          break;
         }

         default:
          errstr<<"Unknown shape in RAW coord subspace ignored"
                << Sas::message(Msg::App, Msg::Sparse);
                                
        } // End of shape case statement

       /* Add shape to output component */
        c.push_back(s);

      } // End of this component

     /* Push back and reset the component */
      outRegion.push_back(c);

      errstr<<"Added component: "<<comp
               << Sas::message(Msg::App, Msg::Verbose);
   }
   RETURN outRegion;
}
DataSubspace * RegionFunctions::detDSS() const
{
  PROC;

  RETURN _dssextra;
}

/* Method to convert all circles to polygons within a region object */
GeoRegion 
RegionFunctions::regionToPoly(GeoRegion inRegion) const
{
  PROC;

   GeoRegion outRegion;

   for (unsigned comp=0; comp<inRegion.size(); comp++) {

      Component c; 

      for (unsigned element=0; element<inRegion[comp].size(); element++) {

        errstr<<"SHAPE is: "<<inRegion[comp][element]->type()
	       << Sas::message(Msg::App, Msg::Verbose);

      /* Copy element to output component, converting to polygons if necc. */

        vector<FShape *> newEls = shapeToPoly(inRegion[comp][element]);

        for (unsigned e=0; e<newEls.size(); e++) {
            c.push_back(newEls[e]);
        }

      }   // End of this component

     /* Push back and reset the component */
      outRegion.push_back(c);

      errstr<<"Added component: "<<comp
	       << Sas::message(Msg::App, Msg::Verbose);
   }

   RETURN outRegion;
}

/* Method: shapeToPoly(FShape element, Component c)
   Purpose: converts a DSS shape into polygons and adds to component
*/
vector<FShape *> RegionFunctions::shapeToPoly(FShape * element) const
{
  PROC;

 /* Pass the shape to the converter. If an annulus, convert into two 
    shapes */

    vector<FShape *> shapes;
    FShape * shape1;
    FShape * shape2;

    switch(element->type())
    {

     /* Convert annular shapes into two shapes */
      case ShapeBasics::ANNULUS :
      case ShapeBasics::ELLIPTANNULUS :
      {
         float cX = element->pointsX()[0];
         float cY = element->pointsY()[0];
         float a = element->radii()[0];
         float b = element->radii()[1];

         errstr<<"Ann centre: "<<cX<<" "<<cY
               <<"Radii: "<<a<<" "<<b
	       << Sas::message(Msg::App, Msg::Verbose);

         if (element->type() == ShapeBasics::ANNULUS) {
            shape1 = new FCircle(FPoint(cX, cY), a, ! element->included());
            shape2 = new FCircle(FPoint(cX, cY), b,  element->included());
         }
         else { // ELLIPTANNULUS:
            float c = element->radii()[2];
            float d = element->radii()[3];

            errstr<<"EllipAnn Radii: "<<c<<" "<<d
	       << Sas::message(Msg::App, Msg::Verbose);

            shape1 = new FEllipse(FPoint(cX, cY), a, b, ! element->included());
            shape2 = new FEllipse(FPoint(cX, cY), c, d, element->included());
         }

         shapes.push_back(shapeToPolyConvert(shape1));
         shapes.push_back(shapeToPolyConvert(shape2));

         break;
      }
      default:
      {
       /* Otherwise convert the single shape */
        shapes.push_back(shapeToPolyConvert(element));
      }
    }
    RETURN shapes;
}

/* Method: shapeToPolyConvert(FShape element)
   Purpose: converts a DSS shape into a polygon
*/
FShape * RegionFunctions::shapeToPolyConvert(FShape * element) const
{
  PROC;

   unsigned Maxpts = 20; // Maximum number of sides in the polygon
   unsigned Npts = 0;
   double angle = 0.0;
   double rotang = 0.0;  // The rotation angle (counter-clockwise !)
   double toang = 0.0;
   vector<float> x(Maxpts), y(Maxpts);

  /* Get centre of the shape */
    float cX = element->pointsX()[0];
    float cY = element->pointsY()[0];
    float trX = 0.0;
    float trY = 0.0;
    float a = 0.0;
    float b = 0.0;
    double degtorad = 3.14159265 / 180.0;
    float theta=0.0;
    float z=0.0;

  /* Get top left, top right, radii and rotation angles from the 
     input element. */ 

   switch(element->type())
     {
      case ShapeBasics::RECTANGLE :
      case ShapeBasics::ROTRECTANGLE :
       {

        // Get top right
         trX = element->pointsX()[1];
         trY = element->pointsY()[1];
         rotang = element->rotAngles()[0] * degtorad;

         a = trX - cX;
         b = trY - cY;

         break;
       }
      case ShapeBasics::BOX :
      case ShapeBasics::ROTBOX :
      case ShapeBasics::ELLIPSE :
      case ShapeBasics::RHOMBUS:
      case ShapeBasics::ROTRHOMBUS:
       {
         a = element->radii()[0];
         b = element->radii()[1];
         rotang = element->rotAngles()[0] * degtorad;

         break;
       }
      case ShapeBasics::SECTOR :
       {
         rotang = element->rotAngles()[0] * degtorad;
         toang = element->rotAngles()[1] * degtorad;

         break;
       }

      case ShapeBasics::CIRCLE :
       {
         a = element->radii()[0];
         b = a;

         break;

      case ShapeBasics::POLYGON :
       { break; }

       }
      default:
              errstr<<"Source region shape cannot be processed"
	       << Sas::error("ShapeNotHandled");
     }

     errstr<<"Point 0: "<<cX<<" "<<cY
     <<"Radii: "<<a<<" "<<b
     <<"Angles "<<rotang<<" "<<toang
	 << Sas::message(Msg::App, Msg::Verbose);

   /* Go round the shape setting points */
   switch(element->type())
     {
      case ShapeBasics::BOX :
      case ShapeBasics::ROTBOX :
        {
       // The radii are actually diameters so divide by 2
           a = a / 2;
           b = b / 2;

           theta = atan(a/b);
           z = sqrt(b*b + a*a);

             Npts = 4;
             x[0] = cX - z * sin(theta-rotang);
             y[0] = cY - z * cos(theta-rotang);
             x[1] = cX - z * sin(theta+rotang);
             y[1] = cY + z * cos(theta+rotang);
             x[2] = cX + z * sin(theta-rotang);
             y[2] = cY + z * cos(theta-rotang);
             x[3] = cX + z * sin(theta+rotang);
             y[3] = cY - z * cos(theta+rotang);

             break;
        }
      case ShapeBasics::RHOMBUS:
      case ShapeBasics::ROTRHOMBUS:
        {
             theta=M_PI/2.0;

             Npts = 4;
             x[0] = cX - a * sin(theta-rotang);
             y[0] = cY - a * cos(theta-rotang);
             x[1] = cX - b * cos(theta-rotang);
             y[1] = cY + b * sin(theta-rotang);
             x[2] = cX + a * sin(theta-rotang);
             y[2] = cY + a * cos(theta-rotang);
             x[3] = cX + b * cos(theta-rotang);
             y[3] = cY - b * sin(theta-rotang);

             break;
        }
      case ShapeBasics::RECTANGLE :
      case ShapeBasics::ROTRECTANGLE :
        {
           theta = atan(a/b);
           z = sqrt(b*b + a*a);

             Npts = 4;
             x[0] = cX;
             y[0] = cY;
             x[1] = cX - b * sin(rotang);
             y[1] = cY + b * cos(rotang);
             x[2] = cX + z * sin(theta-rotang);
             y[2] = cY + z * cos(theta-rotang);
             x[3] = cX + a * cos(rotang);
             y[3] = cY + a * sin(rotang);

             break;
        }
      case ShapeBasics::ELLIPSE :
      case ShapeBasics::CIRCLE :
        {
             Npts = Maxpts;
             for (unsigned n=0; n<Npts; n++) {
               angle += 2.0 * M_PI / Npts;
               x[n] = cX + a * sin(angle);
               y[n] = cY + b * cos(angle);
             }
             break;
        }
      case ShapeBasics::SECTOR:
        {
             Npts = Maxpts;

         // Set the centre point
             x[0] = cX;
             y[0] = cY;

         // Set points along the arc
             a=30000.0;

             for (unsigned n=1; n<Npts; n++) {
               angle=rotang + (toang - rotang) * (n-1) / (Npts-1);
               x[n] = cX + a * sin(angle);
               y[n] = cY + a * cos(angle);

             }
             break;
        }
      case ShapeBasics::POLYGON :
        {
            x = element->pointsX(); 		
            y = element->pointsY(); 		
            Npts = x.size();

            break;
        }

      default:
              errstr<<"Wrong shape - report bug"
	       << Sas::error("ShapeNotHandled");
    }

    for (unsigned n=0; n<Npts; n++) {
         errstr<<"adding point: "<<x[n]<<" "<<y[n]<<" "<<theta
	 << Sas::message(Msg::App, Msg::Noisy);
    }
    
   /* Convert points to detector coordinates if neccessary */
    if (_regPosCoords) 
    {
       for (unsigned n=0; n<Npts; n++) {
           SourcePosition s = _coredata->xyToDetxy(x[n], y[n]);
           x[n] = s.xdet();
           y[n] = s.ydet();
               errstr<<"Det point: "<<x[n]<<" "<<y[n]
	       << Sas::message(Msg::App, Msg::Verbose);
       }
    }
   // else if (_regRawCoords)
    //{
     //  for (unsigned n=0; n<Npts; n++) {
      //     CoordPair s = _coredata->varpositionUtils()->RawTodetxy005(x[n],y[n]);
       //    x[n] = s.x();
        //   y[n] = s.y();

         //      errstr<<"Det3 point: "<<x[n]<<" "<<y[n]
          //     << Sas::message(Msg::App, Msg::Verbose);
       // }
    // }

   /* Return the converted shape and whether it is included/excluded */
    RETURN (new FPolygon(x, y, Npts, element->included()));
}

